import json
import warnings
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from pydantic import model_validator

from first_breaks.data_models.dependent import SGYModel
from first_breaks.data_models.independent import (
    F1F2,
    F3F4,
    Clip,
    ConfidenceOptional,
    DefaultModel,
    ExceptionOptional,
    Gain,
    MaximumTime,
    ModelHashOptional,
    Normalize,
    PicksInSamplesOptional,
    TracesPerGather,
    TracesToInverse,
)
from first_breaks.utils.utils import chunk_iterable, multiply_iterable_by

MINIMUM_TRACES_PER_GATHER = 2


class ProcessingParametersException(Exception):
    pass


class Success(DefaultModel):
    success: Optional[bool] = None


class ErrorMessage(DefaultModel):
    error_message: Optional[str] = None


class Task(
    SGYModel,
    TracesPerGather,
    MaximumTime,
    TracesToInverse,
    F1F2,
    F3F4,
    Gain,
    Clip,
    Normalize,
    PicksInSamplesOptional,
    ConfidenceOptional,
    ErrorMessage,
    Success,
    ModelHashOptional,
    ExceptionOptional,
):
    @property
    def maximum_time_sample(self) -> int:
        return self.sgy.ms2index(self.maximum_time)

    @model_validator(mode="after")  # type: ignore
    def align_parameters(self) -> "Task":
        prev_assignment = self.model_config.get("validate_assignment", None)
        self.model_config["validate_assignment"] = False

        # traces_per_gather
        self.traces_per_gather = min(self.sgy.num_traces, self.traces_per_gather)

        # maximum_time
        maximum_time = self.maximum_time
        if maximum_time == 0.0:
            self.maximum_time = self.sgy.max_time_ms
        else:
            index = self.sgy.ms2index(maximum_time)
            if index == 0:
                warnings.warn(
                    "The maximum time is not zero and is less than the duration of one sample, "
                    "so the maximum time will be equal to the length of the trace."
                )
                self.maximum_time = self.sgy.max_time_ms
            else:
                self.maximum_time = min(maximum_time, self.sgy.max_time_ms)

        # traces_to_inverse
        # to python indices
        traces_to_inverse = [tr - 1 for tr in self.traces_to_inverse if tr <= self.traces_per_gather]
        self.traces_to_inverse = sorted(set(traces_to_inverse))

        self.model_config["validate_assignment"] = prev_assignment
        return self  # type: ignore

    def get_gathers_ids(self) -> List[Tuple[int]]:
        return list(chunk_iterable(list(range(self.sgy.shape[1])), self.traces_per_gather))  # type: ignore

    @property
    def num_gathers(self) -> int:
        return len(self.get_gathers_ids())

    @property
    def picks_in_ms(self) -> Optional[List[float]]:
        if self.picks_in_samples is not None:
            picks_in_samples_list = self._check_and_convert_picks()
            return multiply_iterable_by(picks_in_samples_list, self.sgy.dt_ms)  # type: ignore
        else:
            return None

    @property
    def picks_in_mcs(self) -> Optional[List[int]]:
        if self.picks_in_samples is not None:
            picks_in_samples_list = self._check_and_convert_picks()
            return multiply_iterable_by(picks_in_samples_list, self.sgy.dt_mcs, cast_to=int)  # type: ignore
        else:
            return None

    def _check_and_convert_picks(self) -> List[float]:  # type: ignore
        if self.picks_in_samples is None:
            raise ValueError("There are no picks. Put them manually or process the task first")
        if isinstance(self.picks_in_samples, (tuple, list)):
            picks_in_samples = self.picks_in_samples
        elif isinstance(self.picks_in_samples, np.ndarray):
            picks_in_samples = self.picks_in_samples.tolist()
        else:
            raise TypeError("Only 1D sequence can be saved")
        return picks_in_samples  # type: ignore

    def _prepare_output_for_nonbinary_export(self) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        picks_in_samples = self._check_and_convert_picks()
        picks_in_ms = self.sgy.units_converter.index2ms(picks_in_samples, cast_to=float)

        confidence = self.confidence

        is_source_file = isinstance(self.sgy.source, (str, Path))
        if is_source_file:
            source_filename = str(Path(self.sgy.source).name)  # type: ignore
            source_full_name = str(Path(self.sgy.source).resolve())  # type: ignore
        else:
            source_filename = None
            source_full_name = None

        meta = {
            "is_source_file": is_source_file,
            "is_source_ndarray": self.sgy.is_source_ndarray,
            "filename": source_filename,
            "full_name": source_full_name,
            "hash": self.sgy.get_hash(),
            "dt_ms": self.sgy.dt_ms,
            "is_picked_with_model": bool(self.success),
            "model_hash": self.model_hash,
            "traces_per_gather": self.traces_per_gather,
            "maximum_time": self.maximum_time,
            "traces_to_inverse": self.traces_to_inverse,
            "gain": self.gain,
            "clip": self.clip,
        }
        data = {
            "trace": list(range(1, len(picks_in_samples) + 1)),
            "picks_in_samples": picks_in_samples,
            "picks_in_ms": picks_in_ms,
            "confidence": confidence,
        }
        return meta, data

    def export_result_as_sgy(
        self,
        filename: Union[str, Path],
        byte_position: int = 236,
        encoding: str = "I",
        picks_unit: str = "mcs",
    ) -> None:
        picks_in_samples = self._check_and_convert_picks()
        self.sgy.export_sgy_with_picks(
            output_fname=filename,
            picks_in_samples=picks_in_samples,
            encoding=encoding,
            byte_position=byte_position,
            picks_unit=picks_unit,
        )

    def export_result_as_json(self, filename: Union[str, Path]) -> None:
        meta, data = self._prepare_output_for_nonbinary_export()
        Path(filename).parent.mkdir(parents=True, exist_ok=True)
        with open(filename, "w") as fout:
            json.dump({**meta, **data}, fout)

    def export_result_as_txt(self, filename: Union[str, Path]) -> None:
        meta, data = self._prepare_output_for_nonbinary_export()
        Path(filename).parent.mkdir(parents=True, exist_ok=True)
        with open(filename, "w") as fout:
            content = [f"{k}={v}" for k, v in meta.items()]
            data_str = pd.DataFrame(data).to_string(index=False, justify="right")
            content.append(data_str)
            content = "\n".join(content)
            fout.write(content)



from time import perf_counter
from typing import Any, Optional


class Performance:
    def __init__(self, prefix: Optional[str] = None):
        self.start_time: Optional[float] = None
        self.duration: Optional[float] = None
        self.end_time: Optional[float] = None
        self.prefix = "Duration" if prefix is None else prefix

    def __enter__(self) -> None:
        self.start_time = perf_counter()

    def __exit__(self, *excs: Any) -> None:
        self.end_time = perf_counter()
        self.duration = self.end_time - self.start_time
        msg = f"{self.prefix}: {self.duration}"
        print(msg, flush=True)



from typing import Any, Dict, Optional

from PyQt5.QtGui import QDoubleValidator
from PyQt5.QtWidgets import QApplication, QHBoxLayout, QLabel, QLineEdit, QWidget

from first_breaks.desktop.tooltip_widget import (
    HighlightToolTip,
    ShakeToolTip,
    TextToolTip,
)


class QBandFilterWidget(QWidget):
    def __init__(
        self,
        f1: Optional[float] = None,
        f2: Optional[float] = None,
        f3: Optional[float] = None,
        f4: Optional[float] = None,
        margins: Optional[int] = None,
        debug: bool = False,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)

        self.layout = QHBoxLayout()
        if margins is not None:
            self.layout.setContentsMargins(margins, margins, margins, margins)
        self.setLayout(self.layout)

        self.freq_widgets = {}
        for i, f in enumerate([f1, f2, f3, f4], 1):
            label = QLabel(self._get_freq_str(i))
            self.layout.addWidget(label)

            freq_widget = QLineEdit()
            validator = QDoubleValidator()
            validator.setBottom(0.0)
            freq_widget.setValidator(validator)

            self.layout.addWidget(freq_widget)
            self.freq_widgets[i] = freq_widget
            if debug:
                freq_widget.textChanged.connect(self.validate_and_get_values)

    def enable_freqs_fields(self) -> None:
        for v in self.freq_widgets.values():
            v.setEnabled(True)

    def disable_freqs_fields(self) -> None:
        for v in self.freq_widgets.values():
            v.setEnabled(False)

    @staticmethod
    def _get_freq_str(index: int) -> str:
        return f"f<sub>{index}</sub>"

    def get_raw_freqs(self) -> Dict[int, Optional[float]]:
        freqs = {}
        for i in range(1, 5):
            value = self.freq_widgets[i].text()
            value = float(value) if value else None
            freqs[i] = value
        return freqs

    def validate_and_get_values(self) -> Optional[Dict[int, Optional[float]]]:
        freqs = self.get_raw_freqs()

        problematic_widgets = set()
        recommendation_tips = []

        for idx1, idx2 in [[1, 2], [3, 4]]:
            if (freqs[idx1] is None and freqs[idx2] is not None) or (freqs[idx1] is not None and freqs[idx2] is None):
                text = (
                    f"Parameters {self._get_freq_str(idx1)} and {self._get_freq_str(idx2)} must "
                    f"either be both specified or both empty"
                )
                recommendation_tips.append(text)
                problematic_widgets.add(self.freq_widgets[idx1])
                problematic_widgets.add(self.freq_widgets[idx2])
            elif (freqs[idx1] is not None and freqs[idx2] is not None) and (freqs[idx1] >= freqs[idx2]):
                text = f"{self._get_freq_str(idx2)} must be greater than or equal to {self._get_freq_str(idx1)}"
                recommendation_tips.append(text)
                problematic_widgets.add(self.freq_widgets[idx1])
                problematic_widgets.add(self.freq_widgets[idx2])

        if (freqs[2] is not None and freqs[3] is not None) and (freqs[2] >= freqs[3]):
            text = (
                f"{self._get_freq_str(3)} must be greater than or equal to {self._get_freq_str(2)} if both of "
                f"them are not empty"
            )
            recommendation_tips.append(text)
            problematic_widgets.add(self.freq_widgets[2])
            problematic_widgets.add(self.freq_widgets[3])

        if problematic_widgets:
            if len(recommendation_tips) > 1:
                recommendation_tips = [f"<span>&#8226;</span>{tip}" for tip in recommendation_tips]
                recommendations = "<br>".join(recommendation_tips)
            else:
                recommendations = recommendation_tips[0]

            text_changed_signals = [w.textChanged for w in problematic_widgets]
            problematic_widgets_list = list(problematic_widgets)

            TextToolTip(widget=self, text=recommendations, do_until=text_changed_signals)
            ShakeToolTip(widgets=problematic_widgets_list, parent=self, do_until=2000)
            HighlightToolTip(widgets=problematic_widgets_list, parent=self, do_until=text_changed_signals)
            return None
        else:
            return freqs


if __name__ == "__main__":
    app = QApplication([])
    window = QBandFilterWidget()
    window.show()
    app.exec_()



from typing import Optional, Tuple, Union

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import Polygon


def plotseis(
    data: np.ndarray,
    picking: Optional[np.ndarray] = None,
    add_picking: Optional[np.ndarray] = None,
    normalizing: Optional[Union[str, int]] = "indiv",
    clip: float = 0.9,
    ampl: float = 1.0,
    patch: bool = True,
    colorseis: bool = False,
    wiggle: bool = True,
    background: Optional[np.ndarray] = None,
    colorbar: bool = False,
    dt: float = 1.0,
    show: bool = True,
    dpi: int = 300,
    figsize: Tuple[int, int] = (5, 5),
) -> matplotlib.figure.Figure:

    num_time, num_trace = np.shape(data)

    if normalizing == "indiv":
        norm_factor = np.mean(np.abs(data), axis=0)
        norm_factor[np.abs(norm_factor) < 1e-9 * np.max(np.abs(norm_factor))] = 1
    elif normalizing == "entire":
        norm_factor = np.tile(np.mean(np.abs(data)), (1, num_trace))
    elif np.size(normalizing) == 1 and normalizing is not None:
        norm_factor = np.tile(normalizing, (1, num_trace))
    elif np.size(normalizing) == num_trace:
        norm_factor = np.reshape(normalizing, (1, num_trace))
    elif normalizing is None:
        norm_factor = np.ones(data.shape[1])
    else:
        raise ValueError('Wrong value of "normalizing"')

    data = data / norm_factor * ampl

    mask_overflow = np.abs(data) > clip
    data[mask_overflow] = np.sign(data[mask_overflow]) * clip

    data_time = np.tile((np.arange(num_time) + 1)[:, np.newaxis], (1, num_trace)) * dt

    fig, ax = plt.subplots(figsize=figsize)
    fig.set_dpi(dpi)

    plt.xlim((0, num_trace + 1))
    plt.ylim((0, num_time * dt))
    ax.invert_yaxis()
    ax.xaxis.tick_top()

    if wiggle:
        data_to_wiggle = data + (np.arange(num_trace) + 1)[np.newaxis, :]

        ax.plot(data_to_wiggle, data_time, color=(0, 0, 0))

    if colorseis:
        if not (wiggle or patch):
            ax.imshow(
                data,
                aspect="auto",
                interpolation="bilinear",
                alpha=1,
                extent=(1, num_trace, (num_time - 0.5) * dt, -0.5 * dt),
                cmap="gray",
            )
        else:
            ax.imshow(
                data,
                aspect="auto",
                interpolation="bilinear",
                alpha=1,
                extent=(-0.5, num_trace + 2 - 0.5, (num_time - 0.5) * dt, -0.5 * dt),
                cmap="gray",
            )

    if patch:
        data_to_patch = data
        data_to_patch[data_to_patch < 0] = 0

        for k_trace in range(num_trace):
            patch_data = (
                (data_to_patch[:, k_trace] + k_trace + 1)[:, np.newaxis],
                data_time[:, k_trace][:, np.newaxis],
            )
            patch_data = np.hstack(patch_data)

            head = np.array((k_trace + 1, 0))[np.newaxis, :]
            tail = np.array((k_trace + 1, num_time * dt))[np.newaxis, :]
            patch_data = np.vstack((head, patch_data, tail))

            polygon = Polygon(patch_data, closed=True, facecolor="black", edgecolor=None)
            ax.add_patch(polygon)

    if picking is not None:
        picking = np.array(picking)
        ax.plot(np.arange(num_trace) + 1, picking * dt, linewidth=1, color="blue")

    if add_picking is not None:
        add_picking = np.array(add_picking)
        ax.plot(np.arange(num_trace) + 1, add_picking * dt, linewidth=1, color="green")

    if background is not None:
        bg = ax.imshow(
            background,
            aspect="auto",
            extent=(0.5, num_trace + 1 - 0.5, (num_time - 0.5) * dt, -0.5 * dt),
            cmap="YlOrRd",
        )

        if colorbar:
            plt.colorbar(mappable=bg)

    if show:
        plt.show()
    return fig



import numpy as np
import onnxruntime as ort

from first_breaks import is_windows

ONNX_DEVICE2PROVIDER = {"cuda": "CUDAExecutionProvider", "cpu": "CPUExecutionProvider"}


def is_onnx_gpu_version_installed() -> bool:
    try:
        return ONNX_DEVICE2PROVIDER["cuda"] in ort.get_available_providers()
    except Exception:
        return False


def _colorize(txt: str) -> str:
    _blue_code = "\033[94m"
    _reset = "\033[0m"

    return f"{_blue_code}{txt}{_reset}"


GPU_USSAGE_MESSAGE = """
Before using GPU acceleration, check https://developer.nvidia.com/cuda-gpus that your GPU is CUDA compatible.
"""


CUDA_INSTALLATION_MESSAGE = f"""
Install CUDA Toolkit in one of the following ways:
\t1) Using the official website https://developer.nvidia.com/cuda-toolkit-archive. Select one \
of 11.6, 11.7, 11.8 versions. Version 12 also may work, but versions >=11.6 are recommended.
\t2) Using Conda {_colorize('conda install -c anaconda "cudatoolkit>=11.6,<12"')}.
"""


_CUDNN_LINUX_URL = "https://docs.nvidia.com/deeplearning/cudnn/install-guide/index.html#install-linux"
_CUDNN_WINDOWS_URL = "https://docs.nvidia.com/deeplearning/cudnn/install-guide/index.html#install-windows"

CUDNN_INSTALLATION_MESSAGE = f"""
Install cuDNN toolkit in one of the following ways:
\t1) Using official website {_CUDNN_WINDOWS_URL if is_windows() else _CUDNN_LINUX_URL}.
\t2) Using Conda {_colorize('conda install -c anaconda "cudnn=8.2"')}.
"""


_ZLIB_LIBUX_URL = "https://docs.nvidia.com/deeplearning/cudnn/install-guide/index.html#install-zlib-linux"
_ZLIB_WINDOWS_URL = "https://docs.nvidia.com/deeplearning/cudnn/install-guide/index.html#install-zlib-windows"

ZLIB_INSTALLATION_MESSAGE = f"""
Install ZLib {_ZLIB_WINDOWS_URL if is_windows() else _ZLIB_LIBUX_URL}.
"""


FULL_INSTALLATION_MESSAGE = "".join(
    [GPU_USSAGE_MESSAGE, CUDA_INSTALLATION_MESSAGE, CUDNN_INSTALLATION_MESSAGE, ZLIB_INSTALLATION_MESSAGE]
)


def raise_onnx_cuda_init(add_installation_instruction: bool = True) -> None:
    try:
        ort.OrtValue.ortvalue_from_numpy(np.array([0], dtype=np.float32), "cuda", 0)
    except Exception as exc:
        if add_installation_instruction:
            err_message = f"{str(exc)}\n" f"{'_' * 20}\n" f"Recommendations:\n" f"{FULL_INSTALLATION_MESSAGE}"
            raise type(exc)(err_message).with_traceback(exc.__traceback__)
        else:
            raise exc


def is_onnx_cuda_initializable() -> bool:
    try:
        raise_onnx_cuda_init()
        return True
    except Exception:
        return False


# def is_zlib_installed() -> bool:
#     if is_windows():
#         for path in os.environ["PATH"].split(";"):
#             if Path(path) / "zlibwapi.dll":
#                 pass


def is_onnx_cuda_available() -> bool:
    return is_onnx_gpu_version_installed() and is_onnx_cuda_initializable()


ONNX_CUDA_AVAILABLE = is_onnx_cuda_available()


def get_recommended_device() -> str:
    return "cuda" if ONNX_CUDA_AVAILABLE else "cpu"



from first_breaks.data_models.dependent import Device, XAxis
from first_breaks.data_models.independent import (
    F1F2,
    F3F4,
    BatchSize,
    Clip,
    FillBlack,
    Gain,
    InvertX,
    InvertY,
    MaximumTime,
    Normalize,
    PicksColor,
    RegionContourColor,
    RegionContourWidth,
    RegionPolyColor,
    TracesPerGather,
    VSPView,
)


class Defaults(
    Normalize,
    Gain,
    Clip,
    FillBlack,
    PicksColor,
    RegionPolyColor,
    RegionContourColor,
    RegionContourWidth,
    XAxis,
    F1F2,
    F3F4,
    VSPView,
    InvertX,
    InvertY,
    TracesPerGather,
    Device,
    BatchSize,
    MaximumTime,
):
    pass


DEFAULTS = Defaults()



from pathlib import Path
from typing import Optional, Union

from PyQt5.QtCore import QObject, QThreadPool, pyqtBoundSignal, pyqtSignal
from PyQt5.QtWidgets import QLabel, QProgressBar

from first_breaks.data_models.independent import ExceptionOptional
from first_breaks.desktop.settings_processing_widget import PickingSettings
from first_breaks.desktop.threads import CallInThread, PickerQRunnable
from first_breaks.picking.ipicker import IPicker
from first_breaks.picking.picker_onnx import PickerONNX
from first_breaks.picking.task import Task
from first_breaks.sgy.reader import SGY
from first_breaks.utils.utils import remove_unused_kwargs


class NNManager(QObject):
    picking_not_started_error_signal = pyqtSignal(ExceptionOptional)
    picking_started_signal = pyqtSignal()
    picking_finished_signal = pyqtSignal(Task)

    def __init__(
        self,
        status_progress: QProgressBar,
        status_message: QLabel,
        threadpool: QThreadPool,
        interrupt_on: Union[pyqtSignal, pyqtBoundSignal],
    ):
        super().__init__()
        self.status_progress = status_progress
        self.status_message = status_message
        self.threadpool = threadpool
        self.interrupt_on = interrupt_on

        self.picker_class = PickerONNX
        self.picker_kwargs_init = {"show_progressbar": False, "device": "cpu"}

        self.picker: Optional[IPicker] = None

    def init_net(self, weights: Union[str, Path]) -> None:
        task = CallInThread(self.picker_class, model_path=weights, **self.picker_kwargs_init)
        task.signals.result.connect(self._store_picker)
        self.threadpool.start(task)

    def _store_picker(self, picker: PickerONNX) -> None:
        self.picker = picker

    def pick_fb(self, sgy: SGY, settings: PickingSettings) -> None:
        try:
            settings_dict = settings.model_dump()
            change_settings = remove_unused_kwargs(settings_dict, self.picker.change_settings)  # type: ignore
            self.picker.change_settings(**change_settings)

            task_kwargs = remove_unused_kwargs(settings_dict, Task)  # type: ignore
            task = Task(source=sgy, **task_kwargs)

            worker = PickerQRunnable(picker=self.picker, task=task, interrpution_signal=self.interrupt_on)
            worker.signals.started.connect(self.on_start_task)
            worker.signals.result.connect(self.on_result_task)
            worker.signals.progress.connect(self.on_progressbar_task)
            worker.signals.message.connect(self.on_message_task)
            worker.signals.result.connect(self.on_finish_task)
            self.threadpool.start(worker)
            self.picking_started_signal.emit()
        except Exception as e:
            self.picking_not_started_error_signal.emit(ExceptionOptional(exception=e))

    def on_start_task(self) -> None:
        self.status_progress.show()

    def on_message_task(self, message: str) -> None:
        self.status_message.setText(message)

    def on_progressbar_task(self, value: int) -> None:
        self.status_progress.setValue(value)

    def on_finish_task(self) -> None:
        self.status_progress.hide()

    def on_result_task(self, result: Task) -> None:
        self.picking_finished_signal.emit(result)



import traceback
from typing import List, Literal, Optional, Sequence, Tuple, Union

import numpy as np
from pydantic import BaseModel, Field, field_validator

TColor = Union[Tuple[int, int, int, int], Tuple[int, int, int]]
TNormalize = Union[Literal["trace", "gather"], float, int, np.ndarray, None]


class DefaultModel(BaseModel):
    class Config:
        arbitrary_types_allowed = True
        validate_assignment = True
        protected_namespaces = ()


class TracesPerGather(DefaultModel):
    traces_per_gather: int = Field(12, ge=1, description="How we split the sequence of traces into individual gathers")


class MaximumTime(DefaultModel):
    maximum_time: Union[int, float] = Field(
        0, ge=0.0, description="Limits the window for processing along the time axis"
    )


class TracesToInverse(DefaultModel):
    traces_to_inverse: Sequence[int] = Field((), description="Inverse traces amplitudes on the gathers level")

    @field_validator("traces_to_inverse")
    def validate_traces_to_inverse(cls, traces_to_inverse: Sequence[int]) -> Sequence[int]:
        if not all(val >= 0 for val in traces_to_inverse):
            raise ValueError("Elements of `traces_to_inverse` must be greater or equal to 0")
        return traces_to_inverse


class Normalize(DefaultModel):
    normalize: TNormalize = Field(
        "trace",
        description="How to normalize gather",
    )


class Gain(DefaultModel):
    gain: float = Field(1.0, description="Gain value")


class Clip(DefaultModel):
    clip: float = Field(
        0.9,
        gt=0.0,
        description="Clip amplitudes to this value if their absolute value is greater than this number",
    )


class F1F2(DefaultModel):
    f1_f2: Optional[Tuple[float, float]] = Field(None, description="Frequency pair for growing part band filter")

    @field_validator("f1_f2")
    def validate_encoding(cls, v: Optional[Tuple[float, float]]) -> Optional[Tuple[float, float]]:
        if v is not None:
            if v[0] >= v[1]:
                raise ValueError("f2 should be greater than f1")
        return v


class F3F4(DefaultModel):
    f3_f4: Optional[Tuple[float, float]] = Field(None, description="Frequency pair for decreasing part band filter")

    @field_validator("f3_f4")
    def validate_encoding(cls, v: Optional[Tuple[float, float]]) -> Optional[Tuple[float, float]]:
        if v is not None:
            if v[0] >= v[1]:
                raise ValueError("f4 should be greater than f3")
        return v


class FillBlack(DefaultModel):
    fill_black: Optional[Literal["left", "right"]] = Field(
        "left",
        description="Where and how to fill wiggles with black",
    )


class PicksColor(DefaultModel):
    picks_color: TColor = Field((255, 0, 0), description="Color for picks")


class RegionPolyColor(DefaultModel):
    region_poly_color: TColor = Field((100, 100, 100, 50), description="Color of region below maximum time")


class RegionContourColor(DefaultModel):
    region_contour_color: TColor = Field(
        (100, 100, 100),
        description="Color of dashed line which shows maximum time and how file is split into gathers",
    )


class RegionContourWidth(DefaultModel):
    region_contour_width: float = Field(
        0.2,
        description="Width of dashed line which shows maximum time and how file is split into gathers",
    )


class TraceBytePosition(DefaultModel):
    byte_position: int = Field(0, ge=0, lt=240, description="Byte index in trace headers")


class PicksUnit(DefaultModel):
    picks_unit: Literal["ms", "mcs", "sample"] = Field("mcs", description="First breaks / picks unit")


class PicksInSamplesOptional(DefaultModel):
    picks_in_samples: Optional[Union[np.ndarray, Sequence[Union[int]]]] = Field(
        None, description="First breaks presented as samples"
    )


class ConfidenceOptional(DefaultModel):
    confidence: Optional[Union[np.ndarray, Sequence[Union[int, float]]]] = Field(
        None, description="Confidence of first breaks"
    )


class ModelHashOptional(DefaultModel):
    model_hash: Optional[str] = Field(None, description="Hash of model checkpoint")


class PicksValue(DefaultModel):
    picks_value: Union[np.ndarray, List[Union[int, float]], Tuple[Union[int, float], ...]] = Field(
        ...,
        description="Values of first breaks",
    )


class VSPView(DefaultModel):
    vsp_view: bool = Field(
        False, description="Set the view when the vertical axis is the trace number and the horizontal axis is time"
    )


class InvertX(DefaultModel):
    invert_x: bool = Field(False, description="If True, the X-axis values will increase from right to the left")


class InvertY(DefaultModel):
    invert_y: bool = Field(True, description="If True, the Y-axis values will increase from top to bottom")


class BatchSize(DefaultModel):
    batch_size: int = Field(1, ge=1, description="Batch size")


class ExceptionOptional(DefaultModel):
    exception: Optional[Exception] = None

    def get_formatted_traceback(self) -> str:
        formatted_traceback = traceback.format_exception(
            type(self.exception), self.exception, self.exception.__traceback__
        )
        return "".join(formatted_traceback)



from typing import Dict, List, Tuple


class InvalidHeaders(Exception):
    pass


THeadersAttr = List[Tuple[int, str, str]]


class Headers:
    # headers in format [(offset_1, header_name_1, format_1), (offset_2, header_name_2, format_2), ...]
    # offset started from 0
    headers_schema: THeadersAttr
    format2size: Dict[str, int] = {
        "c": 1,
        "b": 1,
        "B": 1,
        "?": 1,
        "h": 2,
        "H": 2,
        "i": 4,
        "I": 4,
        "l": 4,
        "L": 4,
        "q": 8,
        "Q": 8,
        "f": 4,
        "d": 8,
        "s": 1,
    }

    def get_num_bytes(self, fmt: str) -> int:
        tp = fmt[-1]
        if tp not in self.format2size:
            raise InvalidHeaders("Format is not interpretable")
        num = fmt[:-1]
        if num.isdigit():
            num = int(num)  # type: ignore
        else:
            num = 1  # type: ignore
        return self.format2size[tp] * num  # type: ignore

    def validate(self) -> None:
        headers_names = [header_info[1] for header_info in self.headers_schema]
        if not all(isinstance(name, str) for name in headers_names):
            raise InvalidHeaders("Header names must be strings")
        if not len(set(headers_names)) == len(headers_names):
            raise InvalidHeaders("Header names must be unique")
        available_formats = self.format2size.keys()
        if not all(header_info[2][-1] in available_formats for header_info in self.headers_schema):
            raise InvalidHeaders("Some of headers have unavailable formats")

    def fill_offsets_if_empty(self) -> THeadersAttr:
        offsets_is_none = [header_info[0] is None for header_info in self.headers_schema]
        if all(offsets_is_none):
            upd_headers = []
            offset = 0
            for header in self.headers_schema:
                upd_header = (offset, header[1], header[2])
                upd_headers.append(upd_header)
                offset += self.get_num_bytes(header[2])
            return upd_headers
        elif any(offsets_is_none) and not all(offsets_is_none):
            raise InvalidHeaders("To fill offsets, they must all be either empty or filled (no modification)")
        return self.headers_schema


class FileHeaders(Headers):
    dt_name = "dt"
    dt_name_orig = "dt_orig"
    ns_name = "ns"
    ns_name_orig = "ns_orig"
    data_sample_format_name = "data_sample_format"

    def __init__(self) -> None:
        self.headers_schema = [
            (0, "textual_file_header", "3200s"),
            (3200, "job", "i"),
            (3204, "line", "i"),
            (3208, "reel", "i"),
            (3212, "data_trace_per_ensemble", "H"),
            (3214, "auxiliary_trace_per_ensemble", "H"),
            (3216, self.dt_name, "H"),
            (3218, self.dt_name_orig, "H"),
            (3220, self.ns_name, "H"),
            (3222, self.ns_name_orig, "H"),
            (3224, self.data_sample_format_name, "H"),
            (3226, "ensemble_fold", "H"),
            (3228, "trace_sorting", "h"),
            (3230, "vertical_sum_code", "H"),
            (3232, "sweep_frequency_start", "H"),
            (3234, "sweep_frequency_end", "H"),
            (3236, "sweep_length", "H"),
            (3238, "sweep_type", "h"),
            (3240, "sweep_channel", "h"),
            (3242, "sweep_taper_length_start", "H"),
            (3244, "sweep_taper_length_end", "H"),
            (3246, "taper_type", "H"),
            (3248, "correlated_data_traces", "H"),
            (3250, "binary_gain", "H"),
            (3252, "amplitude_recovery_method", "H"),
            (3254, "measurement_system", "H"),
            (3256, "impulse_signal_polarity", "H"),
            (3258, "vibratory_polarity_code", "H"),
            (3260, "unassigned1", "240s"),
            (3500, "segy_format_revision_number", "H"),
            (3502, "fixed_length_trace_flag", "H"),
            (3504, "number_of_textual_headers", "H"),
            (3506, "unassigned2", "94s"),
        ]
        self.validate()


class TraceHeaders(Headers):
    fb_pick_default = "FB_PICK"

    def __init__(self) -> None:
        self.headers_schema = [
            (0, "TRACENO", "i"),
            (4, "trace_sequence_file", "i"),
            (8, "FFID", "i"),
            (12, "CHAN", "i"),
            (16, "SOURCE", "i"),
            (20, "CDP", "i"),
            (24, "SEQNO", "i"),
            (28, "TRC_TYPE", "h"),
            (30, "STACKNT", "h"),
            (32, "TRFOLD", "h"),
            (34, "data_use", "h"),
            (36, "OFFSET", "i"),
            (40, "REC_ELEV", "i"),
            (44, "SOU_ELEV", "i"),
            (48, "DEPTH", "i"),
            (52, "REC_DATUM", "i"),
            (56, "SOU_DATUM", "i"),
            (60, "SOU_H2OD", "i"),
            (64, "REC_H2OD", "i"),
            (68, "elevation_scalar", "h"),
            (70, "source_group_scalar", "h"),
            (72, "SOU_X", "i"),
            (76, "SOU_Y", "i"),
            (80, "REC_X", "i"),
            (84, "REC_Y", "i"),
            (88, "coordinate_units", "h"),
            (90, "weathering_velocity", "h"),
            (92, "subweathering_velocity", "h"),
            (94, "UPHOLE", "h"),
            (96, "REC_UPHOLE", "h"),
            (98, "SOU_STAT", "h"),
            (100, "REC_STAT", "h"),
            (102, "TOT_STAT", "h"),
            (104, "lag_time_a", "h"),
            (106, "lag_time_b", "h"),
            (108, "delay_recording_time", "h"),
            (110, "TLIVE_S", "h"),
            (112, "TFULL_S", "h"),
            (114, "NUMSMP", "H"),
            (116, "DT", "H"),
            (118, "IGAIN", "h"),
            (120, "PREAMP", "h"),
            (122, "EARLYG", "h"),
            (124, "COR_FLAG", "h"),
            (126, "SWEEPFREQSTART", "h"),
            (128, "SWEEPFREQEND", "h"),
            (130, "SWEEPLEN", "h"),
            (132, "SWEEPTYPE", "h"),
            (134, "SWEEPTAPSTART", "h"),
            (136, "SWEEPTAPEND", "h"),
            (138, "SWEEPTAPCODE", "h"),
            (140, "AAXFILT", "h"),
            (142, "AAXSLOP", "h"),
            (144, "FREQXN", "h"),
            (146, "FXNSLOP", "h"),
            (148, "FREQXL", "h"),
            (150, "FREQXH", "h"),
            (152, "FXLSLOP", "h"),
            (154, "FXHSLOP", "h"),
            (156, "YEAR", "h"),
            (158, "DAY", "h"),
            (160, "HOUR", "h"),
            (162, "MINUTE", "h"),
            (164, "SECOND", "h"),
            (166, "time_basic_code", "h"),
            (168, "trace_weighting_factor", "h"),
            (170, "geophone_group_number_roll1", "h"),
            (172, "geophone_group_number_first", "h"),
            (174, "geophone_group_number_last", "h"),
            (176, "gap_size", "h"),
            (178, "over_travel", "h"),
            (180, "CDP_X", "i"),
            (184, "CDP_Y", "i"),
            (188, "ILINE_NO", "i"),
            (192, "XLINE_NO", "i"),
            (196, "shot_point", "i"),
            (200, "shot_point_scalar", "h"),
            (202, "trace_value_measurement", "h"),
            (204, "transduction_constant_mantissa", "i"),
            (208, "transduction_constant_power", "h"),
            (210, "transduction_unit", "h"),
            (212, "trace_identifier", "h"),
            (214, "scalar_trace_header", "h"),
            (216, "source_type", "h"),
            (218, "source_energy_direction_mantissa", "i"),
            (222, "source_energy_direction_exponent", "H"),
            (224, "source_measurement_mantissa", "i"),
            (228, "source_measurement_exponent", "H"),
            (230, "source_measurement_unit", "h"),
            (232, "unassigned1", "i"),
            (236, self.fb_pick_default, "I"),
        ]
        self.validate()
        self.scalar_from2apply = {
            "elevation_scalar": [
                "REC_ELEV",
                "SOU_ELEV",
                "DEPTH",
                "REC_DATUM",
                "SOU_DATUM",
                "SOU_H2OD",
                "REC_H2OD",
            ],
            "source_group_scalar": [
                "SOU_X",
                "SOU_Y",
                "REC_X",
                "REC_Y",
            ],
            "shot_point_scalar": [
                "shot_point",
            ],
            "scalar_trace_header": [
                "UPHOLE",
                "REC_UPHOLE",
                "SOU_STAT",
                "REC_STAT",
                "TOT_STAT",
                "lag_time_a",
                "lag_time_b",
                "delay_recording_time",
                "TLIVE_S",
                "TFULL_S",
            ],
        }
